# -*- coding: utf-8 -*-
"""
Created on Tue Jul 23 19:33:20 2019

@author: lior
"""
import numpy as np
import xarray as xr
import os, re
import ReadIM


def ReadDavis(path,time=0):
    """
    input path for files format from davis tested for im7&vc7
    out put [X Y U V mask]
    valid only for 2d piv cases
    RETURN:   
     in case of images (image type=0):
              X = scaled x-coordinates
              Y = scaled y-coordinates      
              U = scaled image intensities
              v=0
              MASK=0
            in case of 2D vector fields (A.IType = 1,2 or 3):
              X = scaled x-coordinates
               Y = scaled y-coordinates
               U = scaled vx-components of vectors
               V = scaled vy-components of vectors

    """
    #you need to add clear to prevent data leaks
    buff, vatts   =  ReadIM.extra.get_Buffer_andAttributeList(path)
    v_array, buff1 = ReadIM.extra.buffer_as_array(buff)
    nx=buff.nx
    nz=buff.nz
    ny=buff.ny
    #set data range:
    baseRangeX = np.arange(nx)
    baseRangeY = np.arange(ny)
    baseRangeZ = np.arange(nz)
    lhs1 =(baseRangeX+0.5)*buff.vectorGrid*buff.scaleX.factor+buff.scaleX.offset  # x-range
    lhs2 =(baseRangeY+0.5)*buff.vectorGrid*buff.scaleY.factor+buff.scaleY.offset #y-range
    lhs3 =0
    lhs4 =0
    mask =0
    if buff.image_sub_type<=0: #grayvalue image format
        lhs3 =v_array[0,:,:]
        lhs4=v_array[1,:,:]
        Im = xr.DataArray(v_array,dims=('frame','z','x'),coords={'x':lhs1[0,:],'z':lhs2[:,0],'frame':[0,1]})
        data = xr.Dataset({'Im':Im})
        
    elif buff.image_sub_type==2:# simple 2D vector format: (vx,vy)
    	# Calculate vector position and components
        [lhs1,lhs2] = np.meshgrid(lhs1,lhs2)
    #    lhs1=np.transpose(lhs1)
    #    lhs2=np.transpose(lhs2)
        lhs3 = v_array[0,:,:]*buff.scaleI.factor+buff.scaleI.offset
        lhs4 = v_array[1,:,:]*buff.scaleI.factor+buff.scaleI.offset
        if buff.scaleY.factor<0.0:
            lhs4 = -lhs4
        lhs3=lhs3[:,:,np.newaxis]
        lhs4=lhs4[:,:,np.newaxis]
        u = xr.DataArray(lhs3,dims=('z','x','t'),coords={'x':lhs1[0,:],'z':lhs2[:,0],'t':[time]})
        v = xr.DataArray(lhs4,dims=('z','x','t'),coords={'x':lhs1[0,:],'z':lhs2[:,0],'t':[time]})
        data = xr.Dataset({'u': u, 'v': v})
#    	plt.quiver(lhs1,lhs2,lhs3,lhs4);
    elif buff.image_sub_type==3 or buff.image_sub_type==1:
       #normal 2D vector format + peak: sel+4*(vx,vy) (+peak)
    	#Calculate vector position and components
        [lhs1,lhs2] = np.meshgrid(lhs1,lhs2)
        #    lhs1=np.transpose(lhs1)
        #    lhs2=np.transpose(lhs2)
        lhs3 = lhs1*0;
        lhs4 = lhs2*0;
        # Get choice
        maskData = v_array[0,:,:]
        	# Build best vectors from choice field
        for i in range(5):
            mask =  maskData ==(i+1) 
            if (i<4): # get best vectors
                dat = v_array[2*i+1,:,:]
                lhs3[mask] = dat[mask]
                dat = v_array[2*i+2,:,:]
                lhs4[mask] = dat[mask]		
            else:    # get interpolated vectors
                dat =v_array[7,:,:] 
                lhs3[mask] = dat[mask]
                dat =v_array[8,:,:] 
                lhs4[mask] = dat[mask]
        lhs3 = lhs3*buff.scaleI.factor+buff.scaleI.offset
        lhs4 = lhs4*buff.scaleI.factor+buff.scaleI.offset
        #Display vector field
        if buff.scaleY.factor<0.0:
            lhs4 = -1*lhs4
        mask =  maskData ==0
        lhs3=lhs3[:,:,np.newaxis]
        lhs4=lhs4[:,:,np.newaxis]
        maskData=maskData[:,:,np.newaxis]
        u = xr.DataArray(lhs3,dims=('z','x','t'),coords={'x':lhs1[0,:],'z':lhs2[:,0],'t':[time]})
        v = xr.DataArray(lhs4,dims=('z','x','t'),coords={'x':lhs1[0,:],'z':lhs2[:,0],'t':[time]})
        chc = xr.DataArray(maskData,dims=('z','x','t'),coords={'x':lhs1[0,:],'z':lhs2[:,0],'t':[time]})
        data = xr.Dataset({'u': u, 'v': v,'chc':chc})
    data.attrs['Info'] =ReadIM.extra.att2dict(vatts)
    #clean memory
    ReadIM.DestroyBuffer(buff1)
    del(buff1)
    ReadIM.DestroyBuffer(buff)
    del(buff)
    ReadIM.DestroyAttributeListSafe(vatts)
    del(vatts)
    return data

def load_directory(path,basename=''):
    """ 
    load_directory (path)

    Loads all the .VEC files in the directory into a single
    xarray dataset with variables and units added as attributes

    Input: 
        directory : path to the directory with .vec files

    Output:
        data : xarray DataSet with dimensions: x,y,t and 
               data arrays of u,v,
               attributes of variables and units


    See more: loadvec
    """
    files=[f for f in os.listdir(path) if f.endswith('.vc7')]
    variables, units, rows, cols, dt, frame = parse_header(files[0])
    
    data = []
    for i,f in enumerate(files):
        data.append(loadvec(f,rows,cols,variables,units,dt,frame+i-1))
           
    combined = xr.concat(data, dim='t')
    combined.attrs['variables'] = variables
    combined.attrs['units'] = units
    combined.attrs['dt'] = dt
    combined.attrs['files'] = files
    return combined

path='C:\\Users\\lior\\Documents\\ibrrTau\\timeDependedVecMaps'
files=[f for f in os.listdir(path) if f.endswith('.vc7')]
data=ReadDavis(path+'\\'+files[-1])
#x =x.flatten()
#y =y.ravel()
#u =u.ravel()
#v =v.ravel()
#plt.quiver(x,y,u,v)
#plt.imshow(u)
#    u = xr.DataArray(u,dims=('y','x'),coords={'x':x[0,:],'y':y[:,0]})



# test_methods.py
import xarray as xr
from pivpy import io, pivpy
import numpy as np

import os
f1 = 'Run000001.T000.D000.P000.H001.L.vec'
f2 = 'Run000002.T000.D000.P000.H001.L.vec'
path = os.path.join(os.path.dirname(__file__),'data')

_a = io.load_vec(os.path.join(path,f1))
_b = io.load_vec(os.path.join(path,f2))


def test_crop():
    """ tests crop """
    _c = _a.piv.crop([5, 15,-5,-15])
    assert _c.u.shape == (32,32,1)
    
    _c = io.create_sample_dataset()
    _c = _c.sel(x = slice(35,70),y=slice(30,90))
    assert _c.u.shape == (2,2,5) # note the last dimension is preserved

def test_pan():
    """ test a shift by dx,dy using pan method """
    _a = io.load_vec(os.path.join(path,f1))
    _c = _a.piv.pan(1.0,-1.0) # note the use of .piv.
    assert np.allclose(_c.coords['x'][0],1.312480)
    assert np.allclose(_c.coords['y'][0], -1.31248)


def test_mean():
    data = io.create_sample_dataset(10)
    assert np.allclose(data.piv.average.u.median(), 4.5)

def test_vec2scal():
    data = io.create_sample_dataset()
    data.piv.vec2scal()
    # tests that first data['w'] exists and then 
    # the first value is 0.0
    assert data['w'][0,0,0] == 0.0

def test_add():
    data = io.create_sample_dataset()
    tmp = data + data
    assert tmp['u'][0,0,0] == 2.0

def test_subtract():
    """ tests subtraction """
    data = io.create_sample_dataset()
    tmp = data - data
    assert tmp['u'][0,0,0] == 0.0

def test_multiply():
    """ tests subtraction """
    data = io.create_sample_dataset()
    tmp = data * 3.5
    assert tmp['u'][0,0,0] == 3.5

def test_set_get_dt():
    """ tests setting the new dt """
    data = io.create_sample_dataset()
    assert data.attrs['dt'] == 1.0
    assert data.piv.dt == 1.0
    data.piv.set_dt(2.0)
    assert data.attrs['dt'] == 2.0
    
# def test_rotate():
#     """ tests rotation """
#     data = io.create_sample_dataset()
#     data.piv.rotate(90) # rotate by 90 deg
#     assert data['u'][0,0,0] == 2.1 # shall fail
    
def test_vorticity():
    """ tests vorticity estimate """
    
    data = io.create_sample_field()
    data.piv.vorticity()
    assert data['w'][0,0] == 0.0
    
def test_shear():
    """ tests shear estimate """
    
    data = io.create_sample_field()
    data.piv.shear()
    assert data['w'][0,0] == 0.0 
    
def test_vec2scal():
    """ tests vec2scal """
    data = io.create_sample_field()
    data.piv.vec2scal(property='curl')
    data.piv.vec2scal(property='ke')
    assert len(data.attrs['variables']) == 5
    assert data.attrs['variables'][-1] == 'ke'

    data = io.create_sample_dataset()
    data.piv.vec2scal(property='ke')
    data.piv.vec2scal(property='tke') # now defined
    assert data.attrs['variables'][-1] == 'tke'

    
    _a.piv.vec2scal(property='curl')
    


import numpy as np
from pivpy import io, pivpy
import matplotlib.pyplot as plt

import os
fname = 'Run000001.T000.D000.P000.H001.L.vec'
path = os.path.join(os.path.dirname(__file__),'data')


# def test_get_dt():
#     """ test if we get correct delta t """
#     _, _, _, _,dt,_ = io.parse_header(os.path.join(path,fname))
#     assert dt == 2000.


def test_get_frame():
    """ tests the correct frame number """
    _, _, _, _,_,frame = io.parse_header(os.path.join(path,'day2a005003.T000.D000.P003.H001.L.vec'))
    assert frame == 5003
    _, _, _, _,_,frame = io.parse_header(os.path.join(path,'Run000002.T000.D000.P000.H001.L.vec'))
    assert frame == 2
    _, _, _, _,_,frame = io.parse_header(os.path.join(path,'exp1_001_b.vec'))
    assert frame == 1 
    _, _, _, _,_,frame = io.parse_header(os.path.join(path,'exp1_001_b.txt'))
    assert frame == 1   
    

def test_get_units():
    # test vec file with m/s
    lUnits,vUnits,tUnits = io.get_units(os.path.join(path,fname))
    assert lUnits == 'mm'
    assert vUnits == 'm/s'
    assert tUnits == 's'

    # test vec file with pixels/dt
    lUnits,vUnits,tUnits = io.get_units(os.path.join(path,'day2a005000.T000.D000.P003.H001.L.vec'))
    assert lUnits == 'pixel'
    assert vUnits == 'pixel/dt'
    assert tUnits == 'dt'

    # test OpenPIV vec 
    lUnits,vUnits,tUnits = io.get_units(os.path.join(path,'exp1_001_b.vec') )
    assert lUnits is None

def test_load_vec():
    data = io.load_vec(os.path.join(path,fname))
    assert data['u'].shape == (63,63,1)
    assert data['u'][0,0,0] == 0.0
    assert np.allclose(data.coords['x'][0],0.31248)
    assert 't' in data.dims

def test_load_davis():
    data = io.load_vc7(os.path.join(path,'2Ca.VC7'))
    assert data['u'].shape == (57,43,1)
    assert np.allclose(data.u.values[0,0],-0.04354814)
    assert np.allclose(data.coords['x'][-1],193.313795)

def test_loadopenpivtxt():
    data = io.load_vec(os.path.join(path,'exp1_001_b.txt'))
    

def test_load_directory():
    data = io.load_directory(path,basename='Run*',ext='.vec')
    assert np.allclose(data['t'],[0,1,2,3,4])

    data = io.load_directory(path,basename='2*',ext='.VC7')
    assert np.allclose(data['t'],[0,1])

    data = io.load_directory(path=os.path.join(path,'urban_canopy'),basename='B*',ext='.vc7')
    assert np.allclose(data['t'],[0,1,2,3,4])


def test_create_sample_field():
    data = io.create_sample_field(frame=3)
    assert data['t'] == 3
    data = io.create_sample_field(rows=3,cols=7)
    assert(data.x.shape[0] == 7)
    assert(data.y.shape[0] == 3)
    assert data['t'] == 0

def test_create_sample_dataset(n=3):
    data = io.create_sample_dataset(n=n)
    assert data.dims['t'] == 3
    assert np.allclose(data['t'],np.arange(3)) 




import xarray as xr
from pivpy import io, pivpy, graphics
import numpy as np

import os
f1 = 'Run000001.T000.D000.P000.H001.L.vec'
path = os.path.join(os.path.dirname(__file__),'data')


_d = io.load_vec(os.path.join(path,f1))

def test_showscal():
    graphics.showscal(_d, property='ke')
    

def test_quiver():
    graphics.quiver(_d)


def test_xarray_plot():
    _d.piv.vec2scal(property='curl')
    _d['w'].isel(t=0).plot.pcolormesh()

def test_histogram():
    graphics.histogram(_d)

    
    
    
    
    





# -*- coding: utf-8 -*-
"""
Various plots

"""

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation, FFMpegWriter
import xarray as xr
import os

def quiver(data, arrScale = 25.0, threshold = None, nthArr = 1, 
              contourLevels = None, colbar = True, logscale = False,
              aspectratio='equal', colbar_orient = 'vertical', units = None):
    """
    Generates a quiver plot of a 'data' xarray DataArray object (single frame from a dataset)
    Inputs:
        data - xarray DataArray of the type defined in pivpy, one of the frames in the Dataset
            selected by default using .isel(t=0)
        threshold - values above the threshold will be set equal to threshold
        arrScale - use to change arrow scales
        nthArr - use to plot only every nth arrow from the array 
        contourLevels - use to specify the maximum value (abs) of contour plots 
        colbar - True/False wether to generate a colorbar or not
        logscale - if true then colorbar is on log scale
        aspectratio - set auto or equal for the plot's apearence
        colbar_orient - 'horizontal' or 'vertical' orientation of the colorbar (if colbar is True)
    Outputs:
        none
    Usage:
        graphics.quiver(data, arrScale = 0.2, threshold = Inf, n)
    """
    
    data = dataset_to_array(data)
        
    x = data.x
    y = data.y
    u = data.u.T
    v = data.v.T
    
    if units is not None: # replace  units
        lUnits = units[0] # ['m' 'm' 'mm/s' 'mm/s']
        velUnits = units[2]
        tUnits = velUnits.split('/')[1] # make it 's' or 'dt'
    else:
        lUnits = data.attrs['units'][0]
        velUnits = data.attrs['units'][2]
        tUnits = data.attrs['units'][2].split('/')[-1]
    
    
    if threshold is not None:
        data['u'] = xr.where(data['u']>threshold, threshold, data['u'])
        data['v'] = xr.where(data['v']>threshold, threshold, data['v'])
        
    S = np.array(np.sqrt(u**2 + v**2))
    
    fig = plt.get_fignums()
    if len(fig) == 0: # if no figure is open
        fig, ax = plt.subplots() # open a new figure
    else:
        ax = plt.gca()     
    
    if contourLevels is None:
        levels = np.linspace(0, np.max(S.flatten()), 30) # default contour levels up to max of S
    else:
        levels = np.linspace(0, contourLevels, 30)
                
    if logscale:
        c = ax.contourf(x,y,S,alpha=0.8,
                 cmap = plt.get_cmap("Blues"), 
                 levels = levels, norm = plt.colors.LogNorm())
    else:
        c = ax.contourf(x,y,S,alpha=0.8,
                 cmap = plt.get_cmap("Blues"), 
                 levels=levels)
    if colbar:
        cbar = plt.colorbar(c, orientation=colbar_orient)
        cbar.set_label(r'$ V \, (' + velUnits + r')$' )
        
    ax.quiver(x[::nthArr],y[::nthArr],
               u[::nthArr,::nthArr],v[::nthArr,::nthArr],units='width',
               scale = np.max(S*arrScale),headwidth=2)
    ax.set_xlabel('x (' + lUnits + ')')
    ax.set_ylabel('y (' + lUnits + ')')
    ax.set_aspect(aspectratio)       
    return fig,ax

def histogram(data, normed = False):
    """
    this function will plot a normalized histogram of
    the velocity data.
    Input:
        data : xarray DataSet with ['u','v'] attrs['units']
        normed : (optional) default is False to present normalized
        histogram
        
    """
    
    u = np.asarray(data.u).flatten()
    v = np.asarray(data.v).flatten()
    
    units = data.attrs['units']
    f,ax = plt.subplots(2)
    
    ax[0].hist(u,bins=np.int(np.sqrt(len(u))*0.5),density=normed)
    ax[0].set_xlabel('u ['+units[2]+']')
    
    ax[1] = plt.subplot2grid((2,1),(1,0))
    ax[1].hist(v,bins=np.int(np.sqrt(len(v)*0.5)),density=normed)
    ax[1].set_xlabel('v ['+units[2]+']')
    plt.tight_layout()
    return f, ax


def contour_plot(data, threshold = None, contourLevels = None, 
                    colbar = True,  logscale = False, aspectration='equal', units=None):
    """ contourf ajusted for the xarray PIV dataset, creates a 
        contour map for the data['w'] property. 
        Input:
            data : xarray PIV DataArray, converted automatically using .isel(t=0)
            threshold : a threshold value, default is None (no data clipping)
            contourLevels : number of contour levels, default is None
            colbar : boolean (default is True) show/hide colorbar 
            logscale : boolean (True is default) create in linear/log scale
            aspectration : string, 'equal' is the default
        
    """

    data = dataset_to_array(data)

    
    if units is not None:
        lUnits = units[0] # ['m' 'm' 'mm/s' 'mm/s']
        # velUnits = units[2]
        # tUnits = velUnits.split('/')[1] # make it 's' or 'dt'
    else:
        # lUnits, velUnits = '', ''
        lUnits = data.attrs['units'][0]
        
    f,ax = plt.subplots()    
    
    if threshold is not None:
        data['w'] = xr.where(data['w']>threshold, threshold, data['w'])
        
    m = np.amax(abs(data['w']))
    n = np.amin(abs(data['w']))
    if contourLevels == None:
        levels = np.linspace(np.min(data['w']),np.max(data['w']), 10)
    else:
        levels = contourLevels # vector of levels to set
        
    if logscale:
        c = ax.contourf(data.x,data.y,np.abs(data['w'].T), levels=levels,
                 cmap = plt.get_cmap('RdYlBu'), norm=plt.colors.LogNorm())
    else:
        c = ax.contourf(data.x,data.y,data['w'].T, levels=levels,
                 cmap = plt.get_cmap('RdYlBu'))
        
    plt.xlabel('x [' + lUnits + ']')
    plt.ylabel('y [' + lUnits + ']')
    if colbar:
        cbar = plt.colorbar(c)
        cbar.set_label(r'$\omega$ [s$^{-1}$]')
    ax.set_aspect(aspectration)
    return f,ax
        
         
def showf(data, variables=None, units=None, fig=None):
    """ 
    showf(data, var, units)
    Arguments:
        data : xarray.DataSet that contains dimensions of t,x,y
               and variables u,v and maybe w (scalar)
    """

    if variables is None:
        xlabel = ' '
        ylabel = ' '
    else:
        xlabel = variables[0]
        ylabel = variables[1]

    if units is not None:
        xlabel += ' ' + units[0]
        ylabel += ' ' + units[1]


    fig = plt.figure(None if fig is None else fig.number)


    for t in data['t']:
        d = data.isel(t=t)
        plt.quiver(d['x'],d['y'],d['u'],d['v'],d['u']**2 + d['v']**2)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.draw()
        plt.pause(0.1)
        
    plt.show()

def showscal(data, property='ke'):
    """ 
    showf(data, var, units)
    Arguments:
        data : xarray.DataSet that contains dimensions of t,x,y
               and a variable w (scalar)
    """
    # fig = plt.figure(None if fig is None else fig.number)
    # import pdb; pdb.set_trace()
    # xlabel = (None if var is None else var[0]) + ' [' + (None if units is None else units[0])+']'
    # ylabel = (None if var is None else var[1]) + ' [' + (None if units is None else units[1])+']'
    
    
    data = data.piv.vec2scal(property=property)
    contour_plot(data)                
        

def animate(data, arrowscale=1, savepath=None):
    """ animates the quiver plot for the dataset (multiple frames) 
    Input:
        data : xarray PIV type of DataSet
        arrowscale : [optional] integer, default is 1
        savepath : [optional] path to save the MP4 animation, default is None
    
    Output:
        if savepath is None, then only an image display of the animation
        if savepath is an existing path, a file named im.mp4 is saved
    
    """    
    X, Y = data.x, data.y
    U, V = data.u[:,:,0], data.v[:,:,0] # first frame
    fig, ax = plt.subplots(1,1)
    M = np.sqrt(U**2 + V**2)
    
    Q = ax.quiver(X[::3,::3], Y[::3,::3], 
                  U[::3,::3], V[::3,::3], M[::3,::3],
                 units='inches', scale=arrowscale)
    
    cb = plt.colorbar(Q)
    
    units = data.attrs['units']
    
    cb.ax.set_ylabel('velocity (' + units[2] + ')')
    
    text = ax.text(0.2,1.05, '1/'+str(len(data.t)), ha='center', va='center',
                   transform=ax.transAxes)
    
    def update_quiver(num,Q,data,text):
        U,V = data.u[:,:,num],data.v[:,:,num]
        
        M = np.sqrt(U[::3,::3]**2 + V[::3,::3]**2)   
        Q.set_UVC(U,V,M)
        text.set_text(str(num+1)+'/'+str(len(data.t)))
        return Q

    anim = FuncAnimation(fig, update_quiver, fargs=(Q,data,text),
                               frames = len(data.t), blit=False)
    mywriter = FFMpegWriter()
    if savepath:
        p = os.getcwd()
        os.chdir(savepath)
        anim.save('im.mp4', writer=mywriter)
        os.chdir(p)
    else: anim.save('im.mp4', writer=mywriter)  
    
    

def dataset_to_array(data,N=0):
    """ converts xarray Dataset to array """
    if 't' in data.dims:
        print('Warning: function for a single frame, using first frame, supply data.isel(t=N)')
        data = data.isel(t=N)
    return data


import os
from setuptools import setup, find_packages

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = "pivpy",
    version = "0.0.05",
    author = "Turbulence Structure Laboratory, Tel Aviv University",
    author_email = "alex.liberzon@gmail.com",
    description = ("Post processing of PIV vector files "
                                   " eventually going into OpenPIV "),
    license = "BSD",
    keywords = "example documentation tutorial",
    url = "http://packages.python.org/pivpy",
    packages=find_packages(),
    test_suite="tests",
    install_requires=[
        'numpy',
        'scipy',
        'xarray',
        'matplotlib',
        'readim',
        'pytest'
    ], 
    # packages=['pivpy', 'tests'],
    long_description=read('README.md'),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "License :: OSI Approved :: BSD License",
    ],
)



# -*- coding: utf-8 -*-
"""
Created on Sun May 24 22:02:49 2015

@author: Ron, Alex
"""
import numpy as np
from scipy.stats import norm
from scipy.ndimage.filters import gaussian_filter
from scipy.ndimage.filters import median_filter
import xarray as xr

""" learn from this example 


import xarray as xr


@xr.register_dataset_accessor('geo')
class GeoAccessor(object):
    def __init__(self, xarray_obj):
        self._obj = xarray_obj
        self._center = None

    @property
    def center(self):
        " Return the geographic center point of this dataset."
        if self._center is None:
            # we can use a cache on our accessor objects, because accessors
            # themselves are cached on instances that access them.
            lon = self._obj.latitude
            lat = self._obj.longitude
            self._center = (float(lon.mean()), float(lat.mean()))
        return self._center

    def plot(self):
        " Plot data on a map."
        return 'plotting!'


    In [1]: ds = xr.Dataset({'longitude': np.linspace(0, 10),
   ...:                  'latitude': np.linspace(0, 20)})
   ...: 

In [2]: ds.geo.center
Out[2]: (10.0, 5.0)

In [3]: ds.geo.plot()
Out[3]: 'plotting!'

"""
@xr.register_dataset_accessor('piv')
class PIVAccessor(object):
    def __init__(self,xarray_obj):
        """
        Arguments:
            data : xarray Dataset:
            x,y,t are coordinates
            u,v,chs are the data arrays

        We add few shortcuts (properties):
            data.piv.average is the time average (data.mean(dim='t'))
            data.piv.dt is the shortcut to get $\Delta t$
            data.piv.vorticity
            data.piv.tke
            data.piv.shear

        and a few methods:
            data.piv.vec2scal()
            data.piv.pan 
            data.piv.rotate

        """
        self._obj = xarray_obj
        self._average = None 
        self._dt = None

    @property
    def average(self):
        """ Return the mean flow field ."""
        if self._average is None: # only first time
            self._average = self._obj.mean(dim='t')
            self._average.attrs = self._obj.attrs # we need units in quiver

        return self._average
    
    def crop(self,crop_vector = [None, None, None, None]):
        """ crop number of rows, cols from either side of the vector fields 
        Input: 
            self : xarray Dataset 
            crop_vector : [xmin,xmax,ymin,ymax] is a list of values crop the data, 
            defaults are None
        Return: 
            same object as the input    
        """
        xmin,xmax,ymin,ymax = crop_vector
        
        xmin = self._obj.x.min() if xmin is None else xmin
        xmax = self._obj.x.max() if xmax is None else xmax
        ymin = self._obj.y.min() if ymin is None else ymin
        ymax = self._obj.y.max() if ymax is None else ymax        
        
        self._obj = self._obj.sel(x=slice(xmin, xmax),y=slice(ymin,ymax))

        return self._obj

        

    def pan(self,dx=0.0,dy=0.0):
        """ moves the field by dx,dy in the same units as x,y """
        self._obj['x'] += dx
        self._obj['y'] += dy
        return self._obj

    def filterf(self):
        """Gaussian filtering of velocity """
        from scipy.ndimage.filters import gaussian_filter as gf
        self._obj['u'] = xr.DataArray(gf(self._obj['u'],1),dims=('x','y'))
        self._obj['v'] = xr.DataArray(gf(self._obj['v'],1),dims=('x','y'))
        return self._obj

    def __add__(self,other):
        """ add two datasets means that we sum up the velocities, assume
        that x,y,t,dt are all identical 
        """
        self._obj['u'] += other._obj['u']
        self._obj['v'] += other._obj['v']
        return self._obj

    def __sub__(self,other):
        """ add two datasets means that we sum up the velocities, assume
        that x,y,t,dt are all identical 
        """
        self._obj['u'] -= other._obj['u']
        self._obj['v'] -= other._obj['v']
        return self._obj
    
    
    def vorticity(self):
        """ calculates vorticity of the data array (at one time instance) and 
        adds it to the attributes
        
        Input: 
            xarray with the variables u,v and dimensions x,y
        
        Output:
            xarray with the estimated vorticity as a scalar field with same dimensions
        
        """
        
        ux,_ = np.gradient(self._obj['u'],self._obj['x'],self._obj['y'],axis=(0,1))
        _,vy = np.gradient(self._obj['v'],self._obj['x'],self._obj['y'],axis=(0,1))
        # self._obj['w'] = xr.DataArray(vy - ux, dims=['x', 'y'])
        self._obj['w'] = xr.DataArray(vy - ux, dims=['x', 'y','t'])
        return self._obj
    
    def shear(self):
        """ calculates shear of the data array (single frame) 
        
        Input: 
            xarray with the variables u,v and dimensions x,y
        
        Output:
            xarray with the estimated shear as a scalar field data['w']
        
        """
        
        ux,_ = np.gradient(self._obj['u'],self._obj['x'],self._obj['y'],axis=(0,1))
        _,vy = np.gradient(self._obj['v'],self._obj['x'],self._obj['y'],axis=(0,1))
        # self._obj['w'] = xr.DataArray(vy - ux, dims=['x', 'y'])
        self._obj['w'] = xr.DataArray(vy + ux, dims=['x', 'y','t'])
        return self._obj

    def acceleration(self):
        """ calculates material derivative or acceleration of the data array (single frame) 
        
        Input: 
            xarray with the variables u,v and dimensions x,y
        
        Output:
            xarray with the estimated acceleration as a scalar field data['w']
        
        """
        ux,uy = np.gradient(self._obj['u'],self._obj['x'],self._obj['y'],axis=(0,1))
        vx,vy = np.gradient(self._obj['v'],self._obj['x'],self._obj['y'],axis=(0,1))
        
        ax = self._obj['u']*ux + self._obj['v']*uy
        ay = self._obj['u']*vx + self._obj['v']*vy

        self._obj['w'] = xr.DataArray(np.sqrt(ax**2+ay**2), dims=['x', 'y','t'])
        return self._obj

    def ke(self):
        """ estimates turbulent kinetic energy """
        self._obj['w'] = (self._obj['u'])**2 + (self._obj['v'])**2
        return self._obj
        
    def tke(self):
        """ estimates turbulent kinetic energy """
        if len(self._obj.t) < 2:
            raise ValueError('TKE is not defined for a single vector field, use .piv.ke()')

        self._obj['w'] = (self._obj['u'] - self._obj['u'].mean(dim='t'))**2 + \
            (self._obj['v'] - self._obj['v'].mean(dim='t'))**2
        return self._obj
        
    def vec2scal(self, property='curl'):
        """ creates a dataset of scalar values on the same 
        dimensions and coordinates as the vector dataset
        Agruments:
            data : xarray.DataSet with u,v on t,x,y grid
        Returns:
            scalar_data : xarray.Dataset w on t,x,y grid
            'w' represents one of the following properties:
                - 'curl' or 'rot' - vorticity

        """
        # replace few common names
        property='vorticity' if property == 'curl' else property
        property = 'tke' if property == 'ken' else property
        
        method_name = str(property)
        method = getattr(self, method_name, lambda: "nothing")
        
        if len(self._obj.attrs['variables']) <= 4: # only x,y,u,v
            self._obj.attrs['variables'].append(property)
        else:
            self._obj.attrs['variables'][-1] = property
            
        return method()

    def __mul__(self,scalar):
        '''
        multiplication of a velocity field by a scalar (simple scaling)
        '''
        self._obj['u'] *= scalar
        self._obj['v'] *= scalar

        return self._obj

    def __div__(self,scalar):
        '''
        multiplication of a velocity field by a scalar (simple scaling)
        '''
        self._obj['u'] /= scalar
        self._obj['v'] /= scalar

        return self._obj

    
    def set_dt(self,dt):
        self._obj.attrs['dt'] = dt
        
    def set_tUnits(self,tUnits):
        self._obj.attrs['tUnits'] = tUnits
        
    def rotate(self,theta=0.0):
        """ 
        use this method in order to rotate the data 
        by theta degrees in the clockwise direction
        in the present form of using xarray with coordinates
        it can only work for the cases with equal size along
        x and y 
        """
        
        theta = theta/360.0*2*np.pi
        
        xi = self._obj.x*np.cos(theta) + self._obj.y*np.sin(theta)
        eta = self._obj.y*np.cos(theta) - self._obj.x*np.sin(theta)
        Uxi = self._obj.u*np.cos(theta) + self._obj.v*np.sin(theta)
        Ueta = self._obj.v*np.cos(theta) - self._obj.u*np.sin(theta)
        
        self._obj['x'] = xi
        self._obj['y'] = eta
        self._obj['u'] = Uxi
        self._obj['v'] = Ueta
        
        
        if 'theta' in self._obj:
            self._obj['theta'] += theta
        else:
            self._obj['theta'] = theta
            
        return self._obj
        
    @property
    def dt(self):
        """ receives the dt from the set """
        if self._dt is None:
            self._dt = self._obj.attrs['dt']
        return self._dt
    
        

    # @property
    # def vel_units(self):
    #     " Return the geographic center point of this dataset."
    #     if self._vel_units is None:
    #         self._vel_units = self._obj.attrs.l_units + '/' + self._obj.attrs.t_units 
    #     return self._vel_units
        


# -*- coding: utf-8 -*-

"""
Contains functions for reading flow fields in various formats
"""

import numpy as np
import xarray as xr
from glob import glob
import os, re
import ReadIM




def create_sample_field(rows=5,cols=8,frame = 0):
    """ creates a sample dataset for the tests """

    x  = np.linspace(32.,128.,cols)
    y = np.linspace(16.,128.,rows)

    xm,ym = np.meshgrid(x,y)

    u = np.ones_like(xm.T) + np.linspace(0.0,7.0,rows)
    v = np.zeros_like(ym.T) + np.random.rand(cols,1)-.5

    u = u[:,:,np.newaxis]
    v = v[:,:,np.newaxis]
    chc = np.ones_like(u)

    # plt.quiver(xm.T,ym.T,u,v)



    u = xr.DataArray(u,dims=('x','y','t'),coords={'x':x,'y':y,'t':[frame]})
    v = xr.DataArray(v,dims=('x','y','t'),coords={'x':x,'y':y,'t':[frame]})
    chc = xr.DataArray(chc,dims=('x','y','t'),coords={'x':x,'y':y,'t':[frame]})
    
    data = xr.Dataset({'u': u, 'v': v,'chc':chc})

    data.attrs['variables'] = ['x','y','u','v']
    data.attrs['units'] = ['pix','pix','pix/dt','pix/dt']  
    data.attrs['dt'] = 1.0
    data.attrs['files'] = ''

    return data


def create_sample_dataset(n = 5):
    """ using create_sample_field that has random part in it, create
    a sample dataset of length 'n' """


    data = []
    for i in range(n):
        data.append(create_sample_field(frame=i))
           
    
    combined = xr.concat(data, dim='t')
    combined.attrs['variables'] = ['x','y','u','v']
    combined.attrs['units'] = ['pix','pix','pix/dt','pix/dt']  
    combined.attrs['dt'] = 1.0
    combined.attrs['files'] = ''

    return combined

def from_arrays(x,y,u,v,mask):
    """
        from_arrays(x,y,u,v,mask,frame=0)
        creates an xArray Dataset from 5 two-dimensional Numpy arrays of x,y,u,v and mask

        Input:
            x,y,u,v,mask = Numpy floating arrays, all the same size
        Output:
            data is a xAarray Dataset, see xarray for help 
    """
    # create data structure of appropriate size
    data = create_sample_field(rows=x.shape[0],cols=x.shape[1])
    # assign arrays
    data['x'] = x[0,:]
    data['y'] = y[:,0]
    data['u'] = xr.DataArray(u.T[:,:,np.newaxis],dims=('x','y','t'))
    data['v'] = xr.DataArray(v.T[:,:,np.newaxis],dims=('x','y','t'))
    data['chc'] = xr.DataArray(mask.T[:,:,np.newaxis],dims=('x','y','t'))
    
    return data
        

def load_vec(filename, rows=None, cols=None, variables=None, units=None, dt=None, frame=0):
    """
        load_vec(filename,rows=rows,cols=cols)
        Loads the VEC file (TECPLOT format by TSI Inc.), OpenPIV VEC or TXT formats
        Arguments:
            filename : file name, expected to have a header and 5 columns
            rows, cols : number of rows and columns of a vector field,
            if None, None, then parse_header is called to infer the number
            written in the header
            dt : time interval (default is None)
            frame : frame or time marker (default is None)
        Output:
            data is a xAarray Dataset, see xarray for help 
    """
    if rows is None or cols is None:
        variables,units,rows,cols, dt, frame = parse_header(filename)
    

    if rows is None: # means no headers
        d = np.loadtxt(filename,usecols=(0,1,2,3,4))
        x = np.unique(d[:,0])
        y = np.unique(d[:,1])
        d = d.reshape(len(y),len(x),5).transpose(1,0,2)
    else:
        d = np.loadtxt(filename,skiprows=1,delimiter=',',usecols=(0,1,2,3,4)).reshape(rows,cols,5)
        x = d[:,:,0][0,:]
        y = d[:,:,1][:,0]
        
    u = d[:,:,2]
    v = d[:,:,3]
    chc = d[:,:,4]
    
    # extend dimensions
    u = u[:,:,np.newaxis]
    v = v[:,:,np.newaxis]
    chc = chc[:,:,np.newaxis]

    u = xr.DataArray(u,dims=('x','y','t'),coords={'x':x,'y':y,'t':[frame]})
    v = xr.DataArray(v,dims=('x','y','t'),coords={'x':x,'y':y,'t':[frame]})
    chc = xr.DataArray(chc,dims=('x','y','t'),coords={'x':x,'y':y,'t':[frame]})
    
    data = xr.Dataset({'u': u, 'v': v,'chc':chc})

    data.attrs['variables'] = variables
    data.attrs['units'] = units  
    data.attrs['dt'] = dt
    data.attrs['files'] = filename
    
    return data

def load_directory(path,basename='*',ext='.vec'):
    """ 
    load_directory (path,basename='*', ext='*.vec')

    Loads all the files with the chosen sextension in the directory into a single
    xarray Dataset with variables and units added as attributes

    Input: 
        directory : path to the directory with .vec, .txt or .VC7 files

    Output:
        data : xarray DataSet with dimensions: x,y,t and 
               data arrays of u,v,
               attributes of variables and units


    See more: load_vec
    """
    files  = sorted(glob(os.path.join(path,basename+ext)))
    data = []

    if ext == '.vec':
        variables, units, rows, cols, dt, frame = parse_header(files[0])

        for i,f in enumerate(files):
            data.append(load_vec(f,rows,cols,variables,units,dt,frame+i-1))

    elif ext.lower() == '.vc7':
        frame = 1
        for i,f in enumerate(files):
            data.append(load_vc7(f,frame+i-1))

    if len(data) > 0:
        combined = xr.concat(data, dim='t')
        combined.attrs['variables'] = data[0].attrs['variables']
        combined.attrs['units'] = data[0].attrs['units']
        combined.attrs['dt'] = data[0].attrs['dt']
        combined.attrs['files'] = files

        return combined

    
def parse_header(filename):
    """ 
    parse_header ( filename)
    Parses header of the file (.vec) to get the variables (typically X,Y,U,V)
    and units (can be m,mm, pix/dt or mm/sec, etc.), and the size of the dataset
    by the number of rows and columns.
    Input:
        filename : complete path of the file to read
    Returns:
        variables : list of strings
        units : list of strings
        rows : number of rows of the dataset
        cols : number of columns of the dataset 
        dt   : time interval between the two PIV frames in microseconds
    """

    # split path from the filename
    fname = os.path.basename(filename)
    # get the number in a filename if it's a .vec file from Insight
    if '.' in fname[:-4]: # day2a005003.T000.D000.P003.H001.L.vec
        frame = int(re.findall(r'\d+',fname.split('.')[0])[-1])
    elif '_' in filename[:-4]:
        frame = int(re.findall(r'\d+',fname.split('_')[1])[-1]) # exp1_001_b.vec, .txt

    with open(filename) as fid:
        header = fid.readline()
    
    # if the file does not have a header, can be from OpenPIV or elsewhere
    # return None 
    if header[:5] != 'TITLE':
        return (['x','y','u','v'],None,None,None,None,frame)

    header_list = header.replace(',',' ').replace('=',' ').replace('"',' ').split()
    
    # get variable names, typically X,Y,U,V
    variables = header_list[3:12][::2]
    
    # get units - this is important if it's mm or m/s 
    units = header_list[4:12][::2]

    # get the size of the PIV grid in rows x cols 
    rows = int(header_list[-5])
    cols = int(header_list[-3])

    # this is also important to know the time interval, dt
    ind1 = header.find('MicrosecondsPerDeltaT')
    dt = float(header[ind1:].split('"')[1])
    
    
    return (variables, units, rows, cols, dt, frame)
        


def get_units(filename):
    """ 
    get_units(filename)

    given a full path name to the .vec file will return the names 
    of length and velocity units fallback option is all None. Uses
    parse_header function, see below.

    """

    lUnits, velUnits, tUnits = None, None, None

    _, units, _, _, _, _ = parse_header(filename)

    if units is None:
        return lUnits, velUnits, tUnits

    lUnits = units[0]
    velUnits = units[2]

    if velUnits == 'pixel':
        velUnits = velUnits+'/dt' # make it similar to m/s

    tUnits = velUnits.split('/')[1] # make it 's' or 'dt'
    
    return lUnits, velUnits, tUnits


def ReadDavis(path, frame=1):
    """
    input path for files format from davis tested for im7&vc7
    out put [X Y U V mask]
    valid only for 2d piv cases
    RETURN:   
     in case of images (image type=0):
              X = scaled x-coordinates
              Y = scaled y-coordinates      
              U = scaled image intensities
              v=0
              MASK=0
            in case of 2D vector fields (A.IType = 1,2 or 3):
              X = scaled x-coordinates
               Y = scaled y-coordinates
               U = scaled vx-components of vectors
               V = scaled vy-components of vectors

    """
    #you need to add clear to prevent data leaks
    buff, vatts   =  ReadIM.extra.get_Buffer_andAttributeList(path)
    v_array, buff1 = ReadIM.extra.buffer_as_array(buff)
    nx=buff.nx
    nz=buff.nz
    ny=buff.ny
    #set data range:
    baseRangeX = np.arange(nx)
    baseRangeY = np.arange(ny)
    baseRangeZ = np.arange(nz)
    lhs1 =(baseRangeX+0.5)*buff.vectorGrid*buff.scaleX.factor+buff.scaleX.offset  # x-range
    lhs2 =(baseRangeY+0.5)*buff.vectorGrid*buff.scaleY.factor+buff.scaleY.offset #y-range
    lhs3 =0
    lhs4 =0
    mask =0
    if buff.image_sub_type<=0: #grayvalue image format
        lhs3 =v_array[frame-1,:,:]
        # Display image
#        plt.figure()
#        plt.imshow(lhs3,extent=[lhs1[0],lhs1[-1],lhs2[-1],lhs2[0]],cmap='gray',vmin=0,vmax=1080)
#        plt.colorbar()
    elif buff.image_sub_type==2:# simple 2D vector format: (vx,vy)
    	# Calculate vector position and components
        [lhs1,lhs2] = np.meshgrid(lhs1,lhs2)
    #    lhs1=np.transpose(lhs1)
    #    lhs2=np.transpose(lhs2)
        lhs3 = v_array[0,:,:]*buff.scaleI.factor+buff.scaleI.offset
        lhs4 = v_array[1,:,:]*buff.scaleI.factor+buff.scaleI.offset
        if buff.scaleY.factor<0.0:
            lhs4 = -lhs4
#    	plt.quiver(lhs1,lhs2,lhs3,lhs4);
    elif buff.image_sub_type==3 or buff.image_sub_type==1:
       #normal 2D vector format + peak: sel+4*(vx,vy) (+peak)
    	#Calculate vector position and components
        [lhs1,lhs2] = np.meshgrid(lhs1,lhs2)
        #    lhs1=np.transpose(lhs1)
        #    lhs2=np.transpose(lhs2)
        lhs3 = lhs1*0
        lhs4 = lhs2*0
        # Get choice
        maskData = v_array[0,:,:]
        	# Build best vectors from choice field
        for i in range(5):
            mask =  maskData ==(i+1) 
            if (i<4): # get best vectors
                dat = v_array[2*i+1,:,:]
                lhs3[mask] = dat[mask]
                dat = v_array[2*i+2,:,:]
                lhs4[mask] = dat[mask]		
            else:    # get interpolated vectors
                dat =v_array[7,:,:] 
                lhs3[mask] = dat[mask]
                dat =v_array[8,:,:] 
                lhs4[mask] = dat[mask]
        lhs3 = lhs3*buff.scaleI.factor+buff.scaleI.offset
        lhs4 = lhs4*buff.scaleI.factor+buff.scaleI.offset
        #Display vector field
        if buff.scaleY.factor<0.0:
            lhs4 = -1*lhs4
        mask =  maskData ==0
    #clean memory
    ReadIM.DestroyBuffer(buff1)
    del(buff1)
    ReadIM.DestroyBuffer(buff)
    del(buff)
    ReadIM.DestroyAttributeListSafe(vatts)
    del(vatts)
    return [lhs1,lhs2,lhs3,lhs4,mask]


def load_vc7(filename,frame=0):
    # read the arrays
    x,y,u,v,mask = ReadDavis(filename)
    data = from_arrays(x,y,u,v,mask)
    data['t'] = np.atleast_1d(frame)
    data.attrs['files'] = filename
    
    return data


